from unittest.mock import MagicMock

import pytest

from prefect.testing.utilities import AsyncMock, prefect_test_harness


@pytest.fixture(autouse=True, scope="session")
def prefect_db():
    with prefect_test_harness():
        yield


@pytest.fixture
def slack_credentials():
    slack_credentials_mock = MagicMock()
    chat_postMessage_mock = AsyncMock()
    chat_postMessage_mock.return_value = MagicMock(data=dict())
    slack_credentials_mock.get_client.return_value = MagicMock(
        chat_postMessage=chat_postMessage_mock
    )
    return slack_credentials_mock


@pytest.fixture
def slack_webhook():
    slack_webhook_mock = MagicMock()
    slack_webhook_mock.get_client.return_value = AsyncMock()
    return slack_webhook_mock



from unittest.mock import AsyncMock

import pytest
from prefect_slack import SlackWebhook
from slack_sdk.webhook.async_client import AsyncWebhookClient, WebhookResponse


@pytest.mark.skipif(
    condition=not hasattr(SlackWebhook, "raise_on_failure"),
    reason="The raise_on_failure option is only implemented in prefect>=2.17.2",
)
async def test_slack_webhook_block_does_not_raise_on_success(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr(
        AsyncWebhookClient,
        "send",
        AsyncMock(
            return_value=WebhookResponse(
                url="http://wherever", status_code=222, body="yay", headers={}
            )
        ),
    )
    block = SlackWebhook(url="http://wherever")

    with block.raise_on_failure():
        await block.notify("hello", "world")


@pytest.mark.skipif(
    condition=not hasattr(SlackWebhook, "raise_on_failure"),
    reason="The raise_on_failure option is only implemented in prefect>=2.17.2",
)
async def test_slack_webhook_block_handles_raise_on_failure(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr(
        AsyncWebhookClient,
        "send",
        AsyncMock(
            return_value=WebhookResponse(
                url="http://wherever", status_code=400, body="woops", headers={}
            )
        ),
    )

    from prefect.blocks.abstract import NotificationError

    block = SlackWebhook(url="http://wherever")

    with pytest.raises(NotificationError, match="Failed to send message: woops"):
        with block.raise_on_failure():
            await block.notify("hello", "world")



from unittest.mock import AsyncMock, MagicMock

import pytest
from prefect_slack import SlackCredentials, SlackWebhook
from slack_sdk.web.async_client import AsyncWebClient
from slack_sdk.webhook.async_client import AsyncWebhookClient
from slack_sdk.webhook.webhook_response import WebhookResponse


def test_slack_credentials():
    assert isinstance(SlackCredentials(token="xoxb-xxxx").get_client(), AsyncWebClient)


def test_slack_webhook():
    assert isinstance(
        SlackWebhook(url="https://hooks.slack.com/xxxx").get_client(),
        AsyncWebhookClient,
    )


@pytest.mark.skipif(
    condition=not hasattr(SlackWebhook, "raise_on_failure"),
    reason="The raise_on_failure option is only implemented in prefect>=2.17.2",
)
async def test_slack_webhook_block_does_not_raise_on_success(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr(
        AsyncWebhookClient,
        "send",
        AsyncMock(
            return_value=WebhookResponse(
                url="http://wherever", status_code=222, body="yay", headers={}
            )
        ),
    )
    block = SlackWebhook(url="http://wherever")

    with block.raise_on_failure():
        await block.notify("hello", "world")


@pytest.mark.skipif(
    condition=not hasattr(SlackWebhook, "raise_on_failure"),
    reason="The raise_on_failure option is only implemented in prefect>=2.17.2",
)
async def test_slack_webhook_block_handles_raise_on_failure(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr(
        AsyncWebhookClient,
        "send",
        AsyncMock(
            return_value=WebhookResponse(
                url="http://wherever", status_code=400, body="woops", headers={}
            )
        ),
    )

    from prefect.blocks.abstract import NotificationError

    block = SlackWebhook(url="http://wherever")

    with pytest.raises(NotificationError, match="Failed to send message: woops"):
        with block.raise_on_failure():
            await block.notify("hello", "world")


def test_slack_webhook_sync_notify(monkeypatch):
    """Test the sync notify path"""
    mock_client = MagicMock()
    mock_client.send.return_value = WebhookResponse(
        url="http://test", status_code=200, body="ok", headers={}
    )

    webhook = SlackWebhook(url="http://test")
    monkeypatch.setattr(webhook, "get_client", MagicMock(return_value=mock_client))

    webhook.notify("test message")
    mock_client.send.assert_called_once_with(text="test message")


async def test_slack_webhook_async_notify(monkeypatch):
    """Test the async notify path"""
    mock_client = MagicMock()
    mock_client.send = AsyncMock(
        return_value=WebhookResponse(
            url="http://test", status_code=200, body="ok", headers={}
        )
    )

    webhook = SlackWebhook(url="http://test")
    monkeypatch.setattr(webhook, "get_client", MagicMock(return_value=mock_client))

    await webhook.notify_async("test message")
    mock_client.send.assert_called_once_with(text="test message")


@pytest.mark.parametrize("message", ["test message 1", "test message 2"])
async def test_slack_webhook_notify_async_dispatch(monkeypatch, message):
    """Test that async_dispatch properly handles both sync and async contexts"""

    mock_response = WebhookResponse(
        url="http://test", status_code=200, body="ok", headers={}
    )

    mock_client = MagicMock()
    mock_client.send = AsyncMock(return_value=mock_response)

    webhook = SlackWebhook(url="http://test")
    monkeypatch.setattr(webhook, "get_client", lambda sync_client=False: mock_client)

    # Test notification
    await webhook.notify(message)
    mock_client.send.assert_called_once_with(text=message)



"""Tasks for sending Slack messages."""

from typing import TYPE_CHECKING, Dict, Optional, Sequence, Union

from prefect import task
from prefect.logging import get_run_logger
from prefect_slack.credentials import SlackCredentials, SlackWebhook

if TYPE_CHECKING:  # pragma: no cover
    import slack_sdk.models.attachments
    import slack_sdk.models.blocks


@task
async def send_chat_message(
    channel: str,
    slack_credentials: SlackCredentials,
    text: Optional[str] = None,
    attachments: Optional[
        Sequence[Union[Dict, "slack_sdk.models.attachments.Attachment"]]
    ] = None,
    slack_blocks: Optional[
        Sequence[Union[Dict, "slack_sdk.models.blocks.Block"]]
    ] = None,
) -> Dict:
    """
    Sends a message to a Slack channel.

    Args:
        channel: The name of the channel in which to post the chat message
            (e.g. #general).
        slack_credentials: Instance of `SlackCredentials` initialized with a Slack
            bot token.
        text: Contents of the message. It's a best practice to always provide a `text`
            argument when posting a message. The `text` argument is used in places where
            content cannot be rendered such as: system push notifications, assistive
            technology such as screen readers, etc.
        attachments: List of objects defining secondary context in the posted Slack
            message. The [Slack API docs](https://api.slack.com/messaging/composing/layouts#building-attachments)
            provide guidance on building attachments.
        slack_blocks: List of objects defining the layout and formatting of the posted
            message. The [Slack API docs](https://api.slack.com/block-kit/building)
            provide guidance on building messages with blocks.

    Returns:
        Dict: Response from the Slack API. Example response structures can be found in
            the [Slack API docs](https://api.slack.com/methods/chat.postMessage#examples).

    Example:
        Post a message at the end of a flow run.

        ```python
        from prefect import flow
        from prefect.context import get_run_context
        from prefect_slack import SlackCredentials
        from prefect_slack.messages import send_chat_message


        @flow
        def example_send_message_flow():
            context = get_run_context()

            # Run other tasks and subflows here

            token = "xoxb-your-bot-token-here"
            send_chat_message(
                slack_credentials=SlackCredentials(token),
                channel="#prefect",
                text=f"Flow run {context.flow_run.name} completed :tada:"
            )

        example_send_message_flow()
        ```
    """  # noqa
    logger = get_run_logger()
    logger.info("Posting chat message to %s", channel)

    client = slack_credentials.get_client()
    result = await client.chat_postMessage(
        channel=channel, text=text, blocks=slack_blocks, attachments=attachments
    )
    return result.data


@task
async def send_incoming_webhook_message(
    slack_webhook: SlackWebhook,
    text: Optional[str] = None,
    attachments: Optional[
        Sequence[Union[Dict, "slack_sdk.models.attachments.Attachment"]]
    ] = None,
    slack_blocks: Optional[
        Sequence[Union[Dict, "slack_sdk.models.blocks.Block"]]
    ] = None,
) -> None:
    """
    Sends a message via an incoming webhook.

    Args:
        slack_webhook: Instance of `SlackWebhook` initialized with a Slack
            webhook URL.
        text: Contents of the message. It's a best practice to always provide a `text`
            argument when posting a message. The `text` argument is used in places where
            content cannot be rendered such as: system push notifications, assistive
            technology such as screen readers, etc.
        attachments: List of objects defining secondary context in the posted Slack
            message. The [Slack API docs](https://api.slack.com/messaging/composing/layouts#building-attachments)
            provide guidance on building attachments.
        slack_blocks: List of objects defining the layout and formatting of the posted
            message. The [Slack API docs](https://api.slack.com/block-kit/building)
            provide guidance on building messages with blocks.

    Example:
        Post a message at the end of a flow run.

        ```python
        from prefect import flow
        from prefect_slack import SlackWebhook
        from prefect_slack.messages import send_incoming_webhook_message


        @flow
        def example_send_message_flow():
            # Run other tasks and subflows here

            webhook_url = "https://hooks.slack.com/XXX"
            send_incoming_webhook_message(
                slack_webhook=SlackWebhook(
                    url=webhook_url
                ),
                text="Warehouse loading flow completed :sparkles:"
            )

        example_send_message_flow()
        ```
    """  # noqa
    logger = get_run_logger()
    logger.info("Posting message to provided webhook")

    client = slack_webhook.get_client()
    await client.send(text=text, attachments=attachments, blocks=slack_blocks)



"""Credential classes use to store Slack credentials."""

from typing import Any, Optional, Union

from pydantic import Field, SecretStr
from slack_sdk.web.async_client import AsyncWebClient
from slack_sdk.webhook.async_client import AsyncWebhookClient
from slack_sdk.webhook.client import WebhookClient

from prefect._internal.compatibility.async_dispatch import async_dispatch
from prefect.blocks.core import Block
from prefect.blocks.notifications import NotificationBlock


class SlackCredentials(Block):
    """
    Block holding Slack credentials for use in tasks and flows.

    Args:
        token: Bot user OAuth token for the Slack app used to perform actions.

    Examples:
        Load stored Slack credentials:
        ```python
        from prefect_slack import SlackCredentials
        slack_credentials_block = SlackCredentials.load("BLOCK_NAME")
        ```

        Get a Slack client:
        ```python
        from prefect_slack import SlackCredentials
        slack_credentials_block = SlackCredentials.load("BLOCK_NAME")
        client = slack_credentials_block.get_client()
        ```
    """  # noqa E501

    _block_type_name = "Slack Credentials"
    _logo_url = "https://cdn.sanity.io/images/3ugk85nk/production/c1965ecbf8704ee1ea20d77786de9a41ce1087d1-500x500.png"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-slack/credentials/#prefect_slack.credentials.SlackCredentials"  # noqa

    token: SecretStr = Field(
        default=...,
        description="Bot user OAuth token for the Slack app used to perform actions.",
    )

    def get_client(self) -> AsyncWebClient:
        """
        Returns an authenticated `AsyncWebClient` to interact with the Slack API.
        """
        return AsyncWebClient(token=self.token.get_secret_value())


async def _notify_async(obj: Any, body: str, subject: Optional[str] = None):
    client = obj.get_client()

    response = await client.send(text=body)

    obj._raise_on_failure(response)


class SlackWebhook(NotificationBlock):
    """
    Block holding a Slack webhook for use in tasks and flows.

    Args:
        url: Slack webhook URL which can be used to send messages
            (e.g. `https://hooks.slack.com/XXX`).

    Examples:
        Load stored Slack webhook:
        ```python
        from prefect_slack import SlackWebhook
        slack_webhook_block = SlackWebhook.load("BLOCK_NAME")
        ```

        Get a Slack webhook client:
        ```python
        from prefect_slack import SlackWebhook
        slack_webhook_block = SlackWebhook.load("BLOCK_NAME")
        client = slack_webhook_block.get_client()
        ```

        Send a notification in Slack:
        ```python
        from prefect_slack import SlackWebhook
        slack_webhook_block = SlackWebhook.load("BLOCK_NAME")
        slack_webhook_block.notify("Hello, world!")
        ```
    """

    _block_type_name = "Slack Incoming Webhook"
    _logo_url = "https://images.ctfassets.net/gm98wzqotmnx/7dkzINU9r6j44giEFuHuUC/85d4cd321ad60c1b1e898bc3fbd28580/5cb480cd5f1b6d3fbadece79.png?h=250"  # noqa
    _documentation_url = "https://prefecthq.github.io/prefect-slack/credentials/#prefect_slack.credentials.SlackWebhook"  # noqa

    url: SecretStr = Field(
        default=...,
        title="Webhook URL",
        description="Slack webhook URL which can be used to send messages.",
        examples=["https://hooks.slack.com/XXX"],
    )

    def get_client(
        self, sync_client: bool = False
    ) -> Union[AsyncWebhookClient, WebhookClient]:
        """
        Returns an authenticated `AsyncWebhookClient` to interact with the configured
        Slack webhook.
        """
        if sync_client:
            return WebhookClient(url=self.url.get_secret_value())
        return AsyncWebhookClient(url=self.url.get_secret_value())

    def _raise_on_failure(self, response: Any):
        # prefect>=2.17.2 added a means for notification blocks to raise errors on
        # failures. This is not available in older versions, so we need to check if the
        # private base class attribute exists before using it.
        if getattr(self, "_raise_on_failure", False):  # pragma: no cover
            try:
                from prefect.blocks.abstract import NotificationError
            except ImportError:
                NotificationError = Exception

            if response.status_code >= 400:
                raise NotificationError(f"Failed to send message: {response.body}")

    async def notify_async(self, body: str, subject: Optional[str] = None):
        """
        Sends a message to the Slack channel.
        """
        await _notify_async(self, body, subject)

    @async_dispatch(_notify_async)  # type: ignore
    def notify(self, body: str, subject: Optional[str] = None):
        """
        Sends a message to the Slack channel.
        """
        client = self.get_client(sync_client=True)

        response = client.send(text=body)

        self._raise_on_failure(response)



from itertools import product

import pytest
from prefect_slack.messages import send_chat_message, send_incoming_webhook_message

from prefect import flow

CHANNELS = ["#random", "#general"]
TEXT = ["hello", "goodbye"]
ATTACHMENTS = [
    None,
    [
        {
            "text": "This is an attachment",
            "id": 1,
            "fallback": "This is an attachment's fallback",
        }
    ],
]
SLACK_BLOCKS = [
    None,
    [
        {
            "type": "section",
            "text": {"type": "mrkdwn", "text": "This is a Slack mrkdwn block"},
        }
    ],
]


@pytest.mark.parametrize(
    ["channel", "text", "attachments", "slack_blocks"],
    product(CHANNELS, TEXT, ATTACHMENTS, SLACK_BLOCKS),
)
async def test_send_chat_message(
    slack_credentials, channel, text, attachments, slack_blocks
):
    @flow
    async def test_flow():
        return await send_chat_message(
            slack_credentials=slack_credentials,
            text=text,
            channel=channel,
            attachments=attachments,
            slack_blocks=slack_blocks,
        )

    await test_flow()
    slack_credentials.get_client().chat_postMessage.assert_called_with(
        text=text, channel=channel, blocks=slack_blocks, attachments=attachments
    )


@pytest.mark.parametrize(
    ["text", "attachments", "slack_blocks"],
    product(TEXT, ATTACHMENTS, SLACK_BLOCKS),
)
async def test_send_incoming_webhook_message(
    slack_webhook, text, attachments, slack_blocks
):
    @flow
    async def test_flow():
        await send_incoming_webhook_message(
            slack_webhook=slack_webhook,
            text=text,
            attachments=attachments,
            slack_blocks=slack_blocks,
        )

    await test_flow()
    slack_webhook.get_client().send.assert_called_with(
        text=text, blocks=slack_blocks, attachments=attachments
    )
